#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>

#define SCREEN_WIDTH 128 // OLED display width, in pixels
#define SCREEN_HEIGHT 32 // OLED display height, in pixels


#define OLED_RESET     -1 // Reset pin # (or -1 if sharing Arduino reset pin)
#define SCREEN_ADDRESS 0x3C ///< See datasheet for Address; 0x3D for 128x64, 0x3C for 128x32
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

#define MAX_SEQ 256
byte seq[MAX_SEQ];
byte seq_len;
byte current_pos;
byte max_score;

byte light_pins[] = {9, 10, 11, 12};
byte reset_pin = 4;
byte button_pins[] = {5, 6, 7, 8};
byte buzzer_pin = 3;
int button_sounds[] = {165, 196, 220, 247};
int sound_duration = 100;
int loss_sound = 100;
int loss_sound_duration = 300;
bool is_playing_game;
bool active_buttons;
bool tag_update_display = true;


volatile int next_sound = 0;
volatile int next_sound_duration;

void set_next_sound(int sound, int duration) {
  next_sound = sound;
  next_sound_duration = duration;
}

void update_display() {
  display.clearDisplay();
  display.setCursor(0, 0);
  char text[16];
  sprintf(text, "%d|%d", current_pos, seq_len);
  display.print(text);
  display.setCursor(0, 16);
  sprintf(text, "max: %d", max_score);
  display.print(text);
  display.display();      // Show text
}

void add_to_sequence() {
  seq[seq_len++] = random(0, 4);
  if(max_score < seq_len)
    max_score = seq_len;
}

volatile unsigned long previous_millis;
bool is_playing_sequence = false;

void start_play_sequence() {
  Serial.println(__func__);
  current_pos = 0;
  previous_millis = millis();
  is_playing_sequence = true;
}

void set_leds(byte leds) {
  leds &= 0x0f;
  PORTB &= ~(0x0f << PB1);
  PORTB |= leds << PB1;
}

void stop_game() {
  Serial.println(__func__);
  current_pos = 0;
  seq_len = 0;
  is_playing_game = false;
  is_playing_sequence = false;
  tag_update_display = true;
  set_leds(0);
}





void press_button(int button) {
  // reset leds
  Serial.print("p button ");
  Serial.println(button);
  if (button == seq[current_pos]) {
    Serial.println("right");
    //pressed the right button
    set_leds(1 << button);
    // play sound
    set_next_sound(button_sounds[button], sound_duration);
    current_pos ++;
    if (current_pos == seq_len) {
      // finished current sequence
      // start next one
      is_playing_game = false;
      add_to_sequence();
      start_play_sequence();
    }
    tag_update_display = true;

  } else {
    //pressed the wrong button
    Serial.println("right");
    set_next_sound(loss_sound, loss_sound_duration);
    stop_game();
  }

}



unsigned long lastDebounceTime = 0;  // the last time the output pin was toggled
unsigned long debounceDelay = 150;    // the debounce time; increase if the output flickers




ISR(PCINT2_vect) {
  // PD4(4), PD5(5), PD6(6), PD7(7)

  if (millis() - lastDebounceTime > debounceDelay) {
    lastDebounceTime = millis();
    Serial.println((PIND), BIN);
    if ((PIND & (1 << PD4)) == 0) {

      //Serial.println("aux");
      //Serial.println((PIND & (1 << PD4)), BIN);
      if (is_playing_game) {
        stop_game();
      } else if (is_playing_sequence) {
        stop_game();
      } else {
        add_to_sequence();
        add_to_sequence();
        add_to_sequence();
        start_play_sequence();
      }
      //reset
    }
    if ((PIND & (1 << PD5)) == 0) {
      //button 0
      //Serial.println("button 0");
      if (is_playing_game)
        press_button(0);
    }
    if ((PIND & (1 << PD6)) == 0) {
      //button 1
      //Serial.println("button 1");
      if (is_playing_game)
        press_button(1);
    }
    if ((PIND & (1 << PD7)) == 0) {
      //button 2
      //Serial.println("button 2");
      if (is_playing_game)
        press_button(2);
    }
  }
}
ISR(PCINT0_vect) {
  // PB0(8)
  if (millis() - lastDebounceTime > debounceDelay) {
    lastDebounceTime = millis();

    if ((PINB & (1 << PB0)) == 0) {
      //button 3
      //Serial.println("button 3");
      if (is_playing_game)
        press_button(3);
    }
  }
}

void setup_pins() {

  // inputs
  DDRB &= ~(1 << PB0);
  DDRD &= ~(1 << PD4) & ~(1 << PD5) & ~(1 << PD6) & ~(1 << PD7);
  //pullup
  PORTB |= (1 << PB0);
  PORTD |= (1 << PD4) | (1 << PD5) | (1 << PD6) | (1 << PD7);
  // outputs
  DDRB |= (1 << PB1) | (1 << PB2) | (1 << PB3) | (1 << PB4);

}

void setup_interrupts() {
  cli();

  // întrerupere de tip pin change
  PCMSK2 |= (1 << PCINT20) |  (1 << PCINT21) | (1 << PCINT22) | (1 << PCINT23); //(PD4, PD5, PD6, PD7)
  PCMSK0 |= (1 << PCINT0); //(PD8)

  // întreruperi de tip pin change (activare vector de întreruperi)
  PCICR |=  (1 << PCIE2) | (1 << PCIE0); // enable the pin change interrupt, set PCIE2 to enable PCMSK2 scan


  sei();
}

void setup_display() {
  if (!display.begin(SSD1306_SWITCHCAPVCC, SCREEN_ADDRESS)) {
    Serial.println(F("SSD1306 allocation failed"));
    for (;;); // Don't proceed, loop forever
  }
  display.display();
  display.clearDisplay();
  display.display();
  display.setTextSize(2); // Draw 2X-scale text
  display.setTextColor(SSD1306_WHITE);

}

void setup_serial() {
  Serial.begin(9600);
}

void setup() {
  // put your setup code here, to run once:
  setup_pins();
  setup_interrupts();
  setup_display();
  setup_serial();
  set_leds(15);
  delay(1000);
  set_leds(0);
  Serial.println("start");
}






void loop() {


  // put your main code here, to run repeatedly:

  // 1. generate sequence
  //    -show light
  //    -make sound
  // 2. wait for button press
  // 2.a. correct:
  //      -increment sequence counter
  //      -show button light
  //      -make button sound
  //      -get next step
  // 2.b. incorrect:
  //      -show all lights
  //      -make fail sound
  //      -stop game
  if (is_playing_sequence) {
    if (millis() - previous_millis > sound_duration * 4) {
      Serial.print("pos: ");
      Serial.println(current_pos);
      Serial.print("len: ");
      Serial.println(seq_len);
      previous_millis = millis();
      if (current_pos == seq_len) {
        current_pos = 0;
        is_playing_sequence = false;
        is_playing_game = true;
        tag_update_display = true;
        set_leds(0);
      } else {
        set_leds(1 << seq[current_pos]);
        set_next_sound(button_sounds[seq[current_pos]], sound_duration);
        current_pos++;
      }

    }


  }

  if (next_sound != 0) {
    noTone(buzzer_pin);
    tone(buzzer_pin, next_sound, next_sound_duration);
    next_sound = 0;
  }

  if (tag_update_display) {
    tag_update_display = false;
    update_display();
  }


}
